@@ -1,5 +1,5 @@
-/* Copyright (C) 1995,1996,1997,1998,1999,2000,2001, 2003, 2004, 2005, 2006,
- *   2007, 2008, 2009, 2011, 2013, 2014 Free Software Foundation, Inc.
+/* Copyright (C) 1995-2001, 2003-2009, 2011, 2013, 2014, 2016-2019
+ *   Free Software Foundation, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
@@ -61,11 +61,7 @@
 #include "libguile/stime.h"
 
 #include <unistd.h>
-
-
-#ifdef HAVE_CLOCK_GETTIME
-# include <time.h>
-#endif
+#include <time.h>				  /* Gnulib-provided */
 
 #include <sys/types.h>
 #include <string.h>
@@ -275,9 +271,7 @@ SCM_DEFINE (scm_current_time, "current-time", 0, 0, 0,
 {
   timet timv;
 
-  SCM_CRITICAL_SECTION_START;
   timv = time (NULL);
-  SCM_CRITICAL_SECTION_END;
   if (timv == -1)
     SCM_MISC_ERROR ("current time not available", SCM_EOL);
   return scm_from_long (timv);
@@ -331,7 +325,8 @@ filltime (struct tm *bd_time, int zoff, const char *zname)
   return result;
 }
 
-static char tzvar[3] = "TZ";
+static const char tzvar[3] = "TZ";
+static scm_i_pthread_mutex_t tz_lock = SCM_I_PTHREAD_MUTEX_INITIALIZER;
 
 /* if zone is set, create a temporary environment with only a TZ
    string.  other threads or interrupt handlers shouldn't be allowed
@@ -345,10 +340,11 @@ setzone (SCM zone, int pos, const char *subr)
   if (!SCM_UNBNDP (zone))
     {
       static char *tmpenv[2];
+      char dummy_buf[1];
       char *buf;
       size_t zone_len;
       
-      zone_len = scm_to_locale_stringbuf (zone, NULL, 0);
+      zone_len = scm_to_locale_stringbuf (zone, dummy_buf, 0);
       buf = scm_malloc (zone_len + sizeof (tzvar) + 1);
       strcpy (buf, tzvar);
       buf[sizeof(tzvar)-1] = '=';
@@ -395,9 +391,11 @@ SCM_DEFINE (scm_localtime, "localtime", 1, 1, 0,
 
   itime = SCM_NUM2LONG (1, time);
 
-  /* deferring interupts is essential since a) setzone may install a temporary
-     environment b) localtime uses a static buffer.  */
-  SCM_CRITICAL_SECTION_START;
+  /* Mutual exclusion is essential since a) setzone may install a
+     temporary environment b) localtime uses a static buffer.  */
+  scm_dynwind_begin (0);
+  scm_i_dynwind_pthread_mutex_lock (&tz_lock);
+
   oldenv = setzone (zone, SCM_ARG2, FUNC_NAME);
 #ifdef LOCALTIME_CACHE
   tzset ();
@@ -412,7 +410,7 @@ SCM_DEFINE (scm_localtime, "localtime", 1, 1, 0,
       const char *ptr;
 
       /* copy zone name before calling gmtime or restoring zone.  */
-#if defined (HAVE_TM_ZONE)
+#if defined (HAVE_STRUCT_TM_TM_ZONE)
       ptr = ltptr->tm_zone;
 #elif defined (HAVE_TZNAME)
       ptr = tzname[ (ltptr->tm_isdst == 1) ? 1 : 0 ];
@@ -450,9 +448,9 @@ SCM_DEFINE (scm_localtime, "localtime", 1, 1, 0,
     zoff += 24 * 60 * 60;
 
   result = filltime (&lt, zoff, zname);
-  SCM_CRITICAL_SECTION_END;
-
   free (zname);
+
+  scm_dynwind_end ();
   return result;
 }
 #undef FUNC_NAME
@@ -483,11 +481,11 @@ SCM_DEFINE (scm_gmtime, "gmtime", 1, 0, 0,
 #if HAVE_GMTIME_R
   bd_time = gmtime_r (&itime, &bd_buf);
 #else
-  SCM_CRITICAL_SECTION_START;
+  scm_i_pthread_mutex_lock (&tz_lock);
   bd_time = gmtime (&itime);
   if (bd_time != NULL)
     bd_buf = *bd_time;
-  SCM_CRITICAL_SECTION_END;
+  scm_i_pthread_mutex_unlock (&tz_lock);
 #endif
   if (bd_time == NULL)
     SCM_SYSERROR;
@@ -521,7 +519,7 @@ bdtime2c (SCM sbd_time, struct tm *lt, int pos, const char *subr)
 #if HAVE_STRUCT_TM_TM_GMTOFF
   lt->tm_gmtoff = - scm_to_int (SCM_SIMPLE_VECTOR_REF (sbd_time, 9));
 #endif
-#ifdef HAVE_TM_ZONE
+#ifdef HAVE_STRUCT_TM_TM_ZONE
   if (scm_is_false (SCM_SIMPLE_VECTOR_REF (sbd_time, 10)))
     lt->tm_zone = NULL;
   else
@@ -556,7 +554,7 @@ SCM_DEFINE (scm_mktime, "mktime", 1, 1, 0,
   scm_dynwind_free ((char *)lt.tm_zone);
 #endif
 
-  scm_dynwind_critical_section (SCM_BOOL_F);
+  scm_i_dynwind_pthread_mutex_lock (&tz_lock);
 
   oldenv = setzone (zone, SCM_ARG2, FUNC_NAME);
 #ifdef LOCALTIME_CACHE
@@ -572,7 +570,7 @@ SCM_DEFINE (scm_mktime, "mktime", 1, 1, 0,
       const char *ptr;
 
       /* copy zone name before calling gmtime or restoring the zone.  */
-#if defined (HAVE_TM_ZONE)
+#if defined (HAVE_STRUCT_TM_TM_ZONE)
       ptr = lt.tm_zone;
 #elif defined (HAVE_TZNAME)
       ptr = tzname[ (lt.tm_isdst == 1) ? 1 : 0 ];
@@ -664,9 +662,9 @@ SCM_DEFINE (scm_strftime, "strftime", 2, 0, 0,
   SCM_VALIDATE_STRING (1, format);
   bdtime2c (stime, &t, SCM_ARG2, FUNC_NAME);
 
-  /* Convert string to UTF-8 so that non-ASCII characters in the
-     format are passed through unchanged.  */
-  fmt = scm_to_utf8_stringn (format, &len);
+  /* Convert the format string to the locale encoding, as the underlying
+     'strftime' C function expects.  */
+  fmt = scm_to_locale_stringn (format, &len);
 
   /* Ugly hack: strftime can return 0 if its buffer is too small,
      but some valid time strings (e.g. "%p") can sometimes produce
@@ -682,7 +680,7 @@ SCM_DEFINE (scm_strftime, "strftime", 2, 0, 0,
 
   tbuf = scm_malloc (size);
   {
-#if !defined (HAVE_TM_ZONE)
+#if !defined (HAVE_STRUCT_TM_TM_ZONE)
     /* it seems the only way to tell non-GNU versions of strftime what
        zone to use (for the %Z format) is to set TZ in the
        environment.  interrupts and thread switching must be deferred
@@ -699,10 +697,10 @@ SCM_DEFINE (scm_strftime, "strftime", 2, 0, 0,
 	   some OSs, e.g., Solaris.  */
 	SCM zone =
 	  scm_string_append (scm_list_2 (zone_spec,
-					 scm_from_locale_string ("0")));
+					 scm_from_utf8_string ("0")));
 
 	have_zone = 1;
-	SCM_CRITICAL_SECTION_START;
+        scm_i_scm_pthread_mutex_lock (&tz_lock);
 	oldenv = setzone (zone, SCM_ARG2, FUNC_NAME);
       }
 #endif
@@ -720,16 +718,16 @@ SCM_DEFINE (scm_strftime, "strftime", 2, 0, 0,
 	tbuf = scm_malloc (size);
       }
 
-#if !defined (HAVE_TM_ZONE)
+#if !defined (HAVE_STRUCT_TM_TM_ZONE)
     if (have_zone)
       {
 	restorezone (zone_spec, oldenv, FUNC_NAME);
-	SCM_CRITICAL_SECTION_END;
+        scm_i_pthread_mutex_unlock (&tz_lock);
       }
 #endif
     }
 
-  result = scm_from_utf8_string (tbuf + 1);
+  result = scm_from_locale_string (tbuf + 1);
   free (tbuf);
   free (myfmt);
 #if HAVE_STRUCT_TM_TM_ZONE
@@ -756,16 +754,16 @@ SCM_DEFINE (scm_strptime, "strptime", 2, 0, 0,
 {
   struct tm t;
   char *fmt, *str, *rest;
-  size_t used_len;
+  SCM used_len;
   long zoff;
 
   SCM_VALIDATE_STRING (1, format);
   SCM_VALIDATE_STRING (2, string);
 
-  /* Convert strings to UTF-8 so that non-ASCII characters are passed
-     through unchanged.  */
-  fmt = scm_to_utf8_string (format);
-  str = scm_to_utf8_string (string);
+  /* Convert strings to the locale encoding, as the underlying
+     'strptime' C function expects.  */
+  fmt = scm_to_locale_string (format);
+  str = scm_to_locale_string (string);
 
   /* initialize the struct tm */
 #define tm_init(field) t.field = 0
@@ -784,11 +782,11 @@ SCM_DEFINE (scm_strptime, "strptime", 2, 0, 0,
 
   /* GNU glibc strptime() "%s" is affected by the current timezone, since it
      reads a UTC time_t value and converts with localtime_r() to set the tm
-     fields, hence the use of SCM_CRITICAL_SECTION_START.  */
+     fields, hence the mutex.  */
   t.tm_isdst = -1;
-  SCM_CRITICAL_SECTION_START;
+  scm_i_pthread_mutex_lock (&tz_lock);
   rest = strptime (str, fmt, &t);
-  SCM_CRITICAL_SECTION_END;
+  scm_i_pthread_mutex_unlock (&tz_lock);
   if (rest == NULL)
     {
       /* POSIX doesn't say strptime sets errno, and on glibc 2.3.2 for
@@ -809,14 +807,14 @@ SCM_DEFINE (scm_strptime, "strptime", 2, 0, 0,
   zoff = 0;
 #endif
 
-  /* Compute the number of UTF-8 characters.  */
-  used_len = u8_strnlen ((scm_t_uint8*) str, rest-str);
+  /* Compute the number of characters parsed.  */
+  used_len = scm_string_length (scm_from_locale_stringn (str, rest-str));
   scm_remember_upto_here_2 (format, string);
   free (str);
   free (fmt);
 
   return scm_cons (filltime (&t, zoff, NULL),
-		   scm_from_signed_integer (used_len));
+                   used_len);
 }
 #undef FUNC_NAME
 #endif /* HAVE_STRPTIME */
@@ -834,12 +832,7 @@ scm_init_stime()
 
 #ifdef HAVE_POSIX_CPUTIME
   {
-    clockid_t dummy;
-    
-    /* Only use the _POSIX_CPUTIME clock if it's going to work across
-       CPUs. */
-    if (clock_getcpuclockid (0, &dummy) == 0 &&
-        clock_gettime (CLOCK_PROCESS_CPUTIME_ID, &posix_run_time_base) == 0)
+    if (clock_gettime (CLOCK_PROCESS_CPUTIME_ID, &posix_run_time_base) == 0)
       get_internal_run_time = get_internal_run_time_posix_timer;
     else
       errno = 0;
